/*
 * Programmer:  Timothy E. Hoff
 * File Name:   config.c
 * Written:     July 5, 1996
 * Modified:
 * Description: This module provides access to config records.
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "sfw.h"
#include "pathnames.h"
#include "sizes.h"
#include "input.h"
#include "error.h"
#include "config.h"

#ifndef _LINT
static char rcsid[] = "@(#) $Id: config.c,v 2.2 1996/07/22 20:42:34 t07de7 Exp $";
#endif

/*
 * Private Functions
 */

int static init_rec(struct cmd_alias *, char *);
int static setfield(struct cmd_alias *, char *);
int static cmplen(char *);
void static bad_attribute(char *);
int static addarg(char *[], char *);
int static addtty(char *[], char *);
int static stob(char *);

struct cmd_alias *getcmdalias(char *alias)
{
    /*
     * This function looks up alias in the config file.  If found,
     * a cmd_alias record is populated and the function returns
     * a pointer to the record.  Otherwise the function returns NULL.
     */

    static struct cmd_alias ca;
    FILE *fp;
    char *label, *attr;

    if ((fp = openfile(SFW_CONFIG, "root")) == NULL) {
        fprintf(stderr, "\nERROR:  Could not open SFW_CONFIG.\n");
        fprintf(stderr, "Contact your system administrator.\n");
        exit(EXIT_FAILURE);
    }

#ifdef _DEBUG
    fprintf(stderr, "DEBUG:  Searching for alias [%s].\n", alias);
#endif
    
    while ((label = nextstanza(fp)) != NULL) {
        if (strcmp(label, alias) == 0) {
            
            init_rec(&ca, alias);
            
            
            while ((attr = getattr(fp)) != NULL) {
                if (setfield(&ca, attr) != 0) {
                    close(fp);
                    return NULL;
                }
            }

            close(fp);
            return &ca;
        } 
#ifdef _DEBUG
          else {
            fprintf(stderr, "DEBUG: nextstanza returned [%s]. \
 Still looking...\n", label);
          }
#endif
    }

    close(fp);
    return NULL;
}


int static init_rec(struct cmd_alias *ca, char *label)
{
    /*
     * Initialize the fields in the cmd_alias record
     */

    int i;
    
    memset(ca->alias, '\0', MAXNAME);
    memset(ca->exec_path, '\0', MAXPATHNAME);
    memset(ca->acl_path, '\0', MAXPATHNAME); 
    memset(ca->log_path, '\0', MAXPATHNAME); 
    memset(ca->environ_path, '\0', MAXPATHNAME);                    
    memset(ca->acl_admin, '\0', MAXLOGIN);        
    memset(ca->new_user, '\0', MAXLOGIN);         
    memset(ca->new_group, '\0', MAXGROUP);        
    ca->req_passwd = 0;                         
    ca->pass_args = 1;                          
    memset(ca->auth_notify, '\0', MAX_ADDRESS); 
    memset(ca->admin_notify, '\0', MAX_ADDRESS);
    memset(ca->md5_signature, '\0', MD5_SIZE);

    for (i=0; i<MAXARGC; i++)
        ca->argv[i] = NULL;

    for (i=0; i<MAXTTY; i++)
        ca->tty_list[i] = NULL;               

    sprintf(ca->alias, "%s", label);
    sprintf(ca->acl_path, "%s/%s.acl", ACL_DIR, label);
    sprintf(ca->log_path, "%s/%s.log", LOG_DIR, label);

    return 0;
}
    
int static setfield(struct cmd_alias *ca, char *attr)
{
    /*
     * Set the struct ca field which corresponds to the keyword portion
     * of attr to value.
     */

    char *keyword, *value;
    char msg[BUFSIZE];
    
    if ((keyword = strtok(attr, "=")) != NULL) {
        value = strtok(NULL, "");
    } else {
        /*
         * This code should never execute because getattr() in login.c
         * is supposed to filter for bad data.
         */
        sprintf(msg, "Invalid attribute [%s] passed to setfield().", attr);
        log_error(ESYNTAX, msg);
        fprintf(stderr, "\nERROR:  Unrecognized attribute in SFW_CONFIG.\n");
        fprintf(stderr, "Contact your system administrator.\n");
        return -1;
    }

    switch (*keyword) {
        case 'a':
            if (strncmp(keyword, "argument", cmplen(keyword)) == 0) {
                if (addarg(ca->argv, value) != 0) {
                    fprintf(stderr, "ERROR:  Invalid argument(s)\n");
                    fprintf(stderr, "Contact your system administrator.\n");
                    return -1;
                }
            } else if (strncmp(keyword, "acl", strlen("acl")) == 0) {
                strncpy(ca->acl_path, value, MAXPATHNAME);
            } else if (strncmp(keyword, "auth_notify", cmplen(keyword)) == 0) {
                strncpy(ca->auth_notify, value, MAX_ADDRESS);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 'e':
            if (strncmp(keyword, "executable", cmplen(keyword)) == 0) {
                strncpy(ca->exec_path, value, MAXPATHNAME);
            } else if (strncmp(keyword, "environment", cmplen(keyword)) == 0) {
                strncpy(ca->environ_path, value, MAXPATHNAME);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 'g':
            if (strncmp(keyword, "group", cmplen(keyword)) == 0) {
                strncpy(ca->new_group, value, MAXGROUP);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;

        case 'l':
            if (strncmp(keyword, "log", strlen("log")) == 0) {
                strncpy(ca->log_path, value, MAXPATHNAME);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 'm':
            if (strncmp(keyword, "md5", strlen("md5")) == 0) {
                strncpy(ca->md5_signature, value, MD5_SIZE);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;

        case 'n':
            if (strncmp(keyword, "notify", cmplen(keyword)) == 0) {
                strncpy(ca->admin_notify, value, MAX_ADDRESS);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;

        case 'o':
            if (strncmp(keyword, "owner", cmplen(keyword)) == 0) {
                strncpy(ca->acl_admin, value, MAXLOGIN);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;

        case 'p':
            if (strncmp(keyword, "pass_args", cmplen(keyword)) == 0) {
                ca->pass_args = stob(value);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;

        case 'r':
            if (strncmp(keyword, "req_passwd", cmplen(keyword)) == 0) {
                ca->req_passwd = stob(value);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 't':
            if (strncmp(keyword, "tty", strlen("tty")) == 0) {
                if (addtty(ca->tty_list, value) != 0) {
                    fprintf(stderr, "ERROR:  Invalid argument(s)\n");
                    fprintf(stderr, "Contact your system administrator.\n");
                    return -1;
                }
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
            
        case 'u':
            if (strncmp(keyword, "user", cmplen(keyword)) == 0) {
                strncpy(ca->new_user, value, MAXLOGIN);
            } else {
                bad_attribute(keyword);
                return -1;
            }
            break;
        
        default:
            bad_attribute(keyword);
            return -1;
    }

    return 0;
}


int static cmplen(char *keyword)
{
    /*
     * Return the number of bytes to be used when comparing
     * keyword with defined fields.
     */

    return (strlen(keyword) > 4 ? strlen(keyword) : 4);
}


void static bad_attribute(char *keyword)
{
    char msg[BUFSIZE];

    sprintf(msg, "Configuration keyword [%s] is not defined.", keyword);
    log_error(ESYNTAX, msg);

    return;
}


int static addarg(char *argv[], char *line)
{
    /*
     * Add a new argument to the vector pointed to by argv.
     * If the vector does not exist create it.  line may
     * contain only one argument or it may contain multiple
     * arguments separated by spaces.  Return 0 if successful
     * otherwise -1.
     */

    static int argc = 0;
    int bytes_free;
    char *ptr, *arg;

#define BYTES_USED (ptr - argv[0])

    if (argv[0] == NULL) {
       if ((argv[0] = calloc(MAXARGV, sizeof(char))) == NULL) {
           fprintf(stderr, "ERROR:  Insufficient memory.\n");
           exit(EXIT_FAILURE);
       }
    }

    if ((arg = strtok(line, " \t\n")) == NULL) {
        return -1;
    }

    do {
        /*
         * Calculate the offset into the buffer to use when copying the
         * next argument.  The next argument should begin one byte past
         * the end of the last argument unless it is the first argument.
         */

        if (argc > 0) {
            ptr = argv[argc - 1] + strlen(argv[argc - 1]) + 1;
        } else {
            ptr = argv[0];
        }

        bytes_free = MAXARGV - BYTES_USED;

    	if (strlen(arg) < bytes_free) {
    	    strncpy(ptr, arg, bytes_free);
    	} else {
    	    log_error(E2BIG, "Arguments exceeded MAXARGV");
            return -1;
    	}

        argv[argc] = ptr;
        if (++argc == (MAXARGC - 1)) {
            log_error(E2BIG, "Too many arguments");
            return -1;
        }
        
    } while ((arg = strtok(NULL, " \t\n")) != NULL);
    
    return 0;
}


int static addtty(char *tty_list[], char *line)
{
    /*
     * Add a new tty to the vector pointed to by tty_list.
     * If the vector does not exist create it.  line may
     * contain only one entry or it may contain multiple
     * entries separated by spaces.  Return 0 if successful
     * otherwise -1.
     */

    static int cnt = 0;
    int bytes_free;
    int bytes_used;
    char *ptr, *tty;

    if (tty_list[0] == NULL) {
       if ((tty_list[0] = calloc(MAXTTYV, sizeof(char))) == NULL) {
           fprintf(stderr, "ERROR:  Insufficient memory.\n");
           exit(EXIT_FAILURE);
       }
    }

    if ((tty = strtok(line, " \t\n")) == NULL) {
        return -1;
    }

    do {
        /*
         * Calculate the offset into the buffer to use when copying the
         * next entry.  The next entry should begin one byte past
         * the end of the last unless it is the first.
         */

        if (cnt > 0) {
            ptr = tty_list[cnt - 1] + strlen(tty_list[cnt - 1]) + 1;
        } else {
            ptr = tty_list[0];
        }

        bytes_used = (ptr - tty_list[0]);
        bytes_free = MAXTTYV - bytes_used;

    	if (strlen(tty) < bytes_free) {
    	    strncpy(ptr, tty, bytes_free);
    	} else {
    	    log_error(E2BIG, "TTY list exceeded buffer limits");
            return -1;
    	}

        tty_list[cnt] = ptr;
        if (++cnt == (MAXTTY - 1)) {
            log_error(E2BIG, "Too many ttys");
            return -1;
        }
        
    } while ((tty = strtok(NULL, " \t\n")) != NULL);
    
    return 0;
}


int static stob(char *str)
{
    /*
     * Convert a string value into a boolean representation.
     * The following strings equate to true:  yes, true, y, and t.
     * The following strings equate to false: no, false, n, and f.
     * Any other value which implicitly be treated as false.
     */

    char buf[32];
    int bc;

    memset(buf, '\0', 32);
    strncpy(buf, str, 31);

    for (bc = (strlen(buf)-1); bc>=0; bc--)
        buf[bc] = tolower(buf[bc]);

    if (strlen(buf) == 1) {
        if (*buf == 'y' || *buf == 't')
            return 1;
        else
            return 0;
    } else {
        if ((strcmp(buf, "true") == 0) || (strcmp(buf, "yes") == 0))
            return 1;
        else
            return 0;
    }
}
