/*
 * Programmer:  Timothy E. Hoff
 * File Name:   log.c
 * Written:     July 5, 1996
 * Modified:
 * Description: This module provides all access to log files and syslog.
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include "pathnames.h"
#include "sizes.h"
#include "error.h"
#include "log.h"
#include "util.h"

#ifndef _LINT
static char rcsid[] = "@(#) $Id: log.c,v 2.1 1996/07/10 18:54:37 t07de7 Exp $";
#endif /* _LINT */

static char *err_fmt = "%c [%d] %s\n      %s";
static char *acc_fmt = "%s %c %s %s %s\n"; 

/*
 * Private Function Declarations
 */

void timed_out(int);

void log_error(int code, char *msg)
{
    /*
     * Append an error record to the log file.  If log file does not exist
     * create it.
     */

    int fd;
    char buf[BUFSIZE];
    time_t tv;
    extern char *err_fmt;
    extern int errcode;

    errcode = code;
    
    /* Format the log entry */
     
    time(&tv);
    sprintf(buf, err_fmt, 'E', code, msg, ctime(&tv));
    
    if ((fd = open(SFW_LOG, O_WRONLY | O_APPEND | O_CREAT,
        S_IRUSR | S_IWUSR | S_IRGRP)) == -1) {
#ifdef _DEBUG
        fprintf(stderr, "DEBUG:  Could not open %s.\n", SFW_LOG);
#endif
        return;
    }

    if (get_lock(fd, F_WRLCK, 0, SEEK_CUR, 0) == 0) {
        if (write(fd, buf, strlen(buf)) == -1) {
#ifdef _DEBUG
            fprintf(stderr, "DEBUG:  Could not write to %s.\n", SFW_LOG);
            fprintf(stderr, "DEBUG:  errno = %d.\n", errno);
#endif
        }
    } else {
        fprintf(stderr, "ERROR:  Required resource is busy.  Try again later.\n");
        exit(EXIT_FAILURE);
    }

    close(fd);      /* close the log and release write lock */
    return;
}


void log_access(char *pathname, char *login, char flag, char **argv)
{
    /*
     * Append a record to the log specified by pathname for access
     * by login.  flag is either a + (plus) or (minus) indicating
     * whether access was granted or denied.  argv is what is passed
     * to exec().
     */

    int fd;
    char buf[BUFSIZE];
    char argl[BUFSIZE];
    char time_stamp[32];
    extern char *acc_fmt;
    char *terminal;
    time_t tv;
    int bc = BUFSIZE;

    /* Format the record */

    memset(buf, '\0', BUFSIZE);
    memset(argl, '\0', BUFSIZE);
    
    (void) time(&tv);
    strncpy(time_stamp, ctime(&tv), sizeof(time_stamp));
    time_stamp[24] = '\0';          /* remove the newline */

    if ((terminal = getctty()) == NULL) {
        terminal = "-";
    }
    
    while (*argv) {
        strncat(argl, *argv, bc);
        bc -= strlen(*argv) + 1;
        strncat(argl, " ", bc);
        argv++;
    }
        
    sprintf(buf, acc_fmt, time_stamp, flag, login, terminal, argl);
    
    if ((fd = open(pathname, O_WRONLY | O_APPEND | O_CREAT,
        S_IRUSR | S_IWUSR | S_IRGRP)) == -1) {
#ifdef _DEBUG
        fprintf(stderr, "DEBUG:  Could not open %s.\n", pathname);
#endif
        sprintf(buf, "open(%s) failed (errno %d)", pathname, errno);
        log_error(EACCES, buf);
        return;
    }

    if (get_lock(fd, F_WRLCK, 0, SEEK_CUR, 0) == 0) {
        if (write(fd, buf, strlen(buf)) == -1) {
#ifdef _DEBUG
            fprintf(stderr, "DEBUG:  Could not write to %s.\n", pathname);
            fprintf(stderr, "DEBUG:  errno = %d.\n", errno);
#endif
            sprintf(buf, "write() failed (errno %d)", errno);
            log_error(EACCES, buf);
            exit(EXIT_FAILURE);
        }
    } else {
        fprintf(stderr, "ERROR:  Required resource is busy.  Try again later.\n");
        exit(EXIT_FAILURE);
    }

    close(fd);      /* close the log and release write lock */
    return;
}
    

int get_lock(int fd, int type, off_t start, short whence, off_t len)
{
    /*
     * Attempt to obtain a lock on a region of the file pointed to
     * by fd.  First, set an alarm for TIMEOUT seconds and then
     * block waiting for the lock.  If the lock returns clear
     * the alarm and return 0.  If the alarm expires before the
     * lock return -1.
     */

    struct flock lock;
    struct sigaction act, oact;
    int nsec, prev_nsec;

    lock.l_type = type;
    lock.l_start = start;
    lock.l_whence = whence;
    lock.l_len = len;

    act.sa_handler = timed_out;

    if (sigaction(SIGALRM, &act, &oact) == -1) {
        fprintf(stderr, "ERROR:  Could not install signal handler.\n");
        exit(EXIT_FAILURE);
    }
    
    prev_nsec = alarm(TIMEOUT);

    if (fcntl(fd, F_SETLKW, &lock) == -1) {
#ifdef _DEBUG        
        if (errno == EINTR) {
            fprintf(stderr, "DEBUG:  get_lock() interupted by a signal.\n");
        } else {
            fprintf(stderr, "DEBUG:  get_lock() failed (errno %d).\n", errno);
        }
#endif            
        return -1;                  /* region not locked */
    }

    nsec = alarm(0);                /* Cancel alarm */

    /*
     * Restore the previous alarm.  If the time left on the alarm was
     * greater than the time get_lock used restore the unused portion.
     * Otherwise pretend that get_lock did not take any time.
     */
     
    if (prev_nsec > 0) {
        if (prev_nsec > (TIMEOUT - nsec))
            (void) alarm(prev_nsec - (TIMEOUT - nsec));
        else
            (void) alarm(prev_nsec);
    }

    return 0;                       /* region was locked */
}


void timed_out(int val)
{
    /*
     * Signal handler for alarm()
     */

    return;
}
