/*
 * Programmer:  Timothy E. Hoff
 * File Name:   md5check.c
 * Written:     July 7, 1996
 * Modified:
 * Description: This program computes a checksum on a file
 *              using the MD5 algorithm from Ron Rivest.
 *              This program is derived from the RSA Data
 *              Security, Inc. MD5 Message-Digest Algorithm.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "md5.h"


#define CHUNKSIZE   64
#define BLOCKSIZE   2048

/*
 * Function Declarations
 */

int md5_check(char *, char *);
int MD5Calc(char *, char *);
void MD5Format(unsigned char *, char *);


int md5_check(char *pathname, char *signature)
{
    /*
     * Calculate the message digest for pathname and compare it to
     * signature.  If the two match return 0 otherwise return -1.
     */

    char md5_buf[33];               /* 16 hex words */

    memset(md5_buf, '\0', sizeof(md5_buf));

    if (MD5Calc(pathname, md5_buf) == 0) {
        if (strcmp(signature, md5_buf) == 0) {
            return 0;
        } else {
            return -1;
        }
    } else {
        return -1;                  /* MD5Calc failed */
    }
}
 

int MD5Calc(char *filename, char *md5_buf)
{
    /*
     * Calculate the MD5 message-digest for filename.  Store the signature
     * in buffer.  Return 0 if successful otherwise -1.
     */

    int fd;
    void *buf;
    int bc;
    extern int errno;
    MD5_CTX context;
    unsigned char digest[16];


    if ((fd = open(filename, O_RDONLY, 0)) == -1) {
        fprintf(stderr, "ERROR (md5check.c):  Could not open %s.\n", filename);
        return -1;
    }
    
    /* Allocate buffer of nBlockSize */

    if ((buf = malloc(BLOCKSIZE)) == NULL) {
        fprintf(stderr, "ERROR (md5check.c):  Out of memory.\n");
        return -1;
    }

    /* Initialize the MD5 structure */

    MD5Init(&context);

    while ((bc = read(fd, buf, BLOCKSIZE)) > 0) {
        MD5Update(&context, buf, bc);
    }

    if (bc == -1) {
        /* I/O Error */
        fprintf(stderr, "ERROR (sum.c):  I/O error.\n");
        return -1;
    }

    MD5Final(digest, &context);
    MD5Format(digest, md5_buf);

    return 0;
}


void MD5Format(unsigned char *digest, char *buf)
{
    /*
     * Print the message digest as a 32-digit string of hexadecimal
     * digits.  This requires that we split unsigned chars into
     * high and lo bits (4-bits each).
     */

    int i;
    unsigned char hi, lo;
    char word[3];

    *buf = '\0';                /* clear buf */
    
    for (i=0; i<16; i++) {
        hi = lo = digest[i];
        hi = (hi >> (7+1-4) & ~(~0 << 4));
        lo = (lo >> (3+1-4) & ~(~0 << 4));
        
        sprintf(word, "%x%x", hi, lo);
        strcat(buf, word);
    }
    
    return;
}



