/*
 * Programmer:  Timothy E. Hoff
 * File Name:   util.c
 * Written:     July 5, 1996
 * Modified:
 * Description: This module provides various utility functions and
 *              routines.
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <string.h>
#include <errno.h>
#include "sizes.h"
#include "util.h"
#include "log.h"

 
uid_t user_to_id(char *user)
{
    /*
     * Lookup user in the password file and return the uid.  If
     * the user is not defined return -1.
     */

    struct passwd *ptr;

    if ((ptr = getpwnam(user)) != NULL) {
        return ptr->pw_uid;
    } else {
        return -1;
    }
}


char *getctty()
{
    /*
     * Return the name of the controlling terminal or NULL if the
     * process is not attached to a terminal.
     */
     
    int fd;
    char *tty;
    
#ifdef _SVR4
    for (fd = STDIN_FILENO; fd <= STDERR_FILENO; fd++) {
        if (isatty(fd)) {
            break;
        }
    }
	if (fd > STDERR_FILENO)
		fd = -1;

#else
	fd = open("/dev/tty", O_RDONLY);
#endif
    
	if (fd != -1) {
        tty = ttyname(fd);          /* returns a pointer to terminal's pathname */
    } else {
		return NULL;
	}

    /* strip off "/dev/" */

    if (strncmp(tty, "/dev/", 5) == 0) {
        tty += strlen("/dev/");
    }
    
    return tty;
}


char *strip_path(char *pathname)
{
    /*
     * Remove the path component from pathname and return a
     * pointer to the non-qualified file.
     */

    char *ch;

    if ((ch = strrchr(pathname, '/')) == NULL) {
        return pathname;
    } else {
        return ++ch;
    }
}

#ifndef _HAVE_SETENV

int setenv(const char *name, const char *value, int rewrite)
{
    /*
     * Set name to value.  If name already exists in the environment
     * and rewrite is non-zero it is replaced otherwise the existing
     * definition remains.
     */

    char *envar;
    static char buf[BUFSIZE];
    char *ptr;

    if (rewrite == 0) {
        if ((envar = getenv(name)) != NULL) {
            return 0;
        }
    }

    sprintf(buf, "%s=%s", name, value);

    if ((ptr = calloc(strlen(buf) + 1, sizeof(char))) == NULL) {
        log_error(E2BIG, "Insufficient memory for environment.");
	return -1;
    } else {
        strcpy(ptr, buf);
    }

    return putenv(ptr);
}

#endif /* _HAVE_SETENV */


#ifndef _HAVE_ACCESSX

int accessx(char *pathname, int mode, int who)
{
    /*
     * Determines the accessibility of pathname.  The
     * mode parameter is a bit-mask containing 0 or more
     * of the following R_OK, W_OK, X_OK or F_OK.  The who
     * parameter specifies a class of users for whom
     * accessibility is checked.  This parameter must be
     * one of the following ACC_SELF, ACC_INVOKER, ACC_OTHERS
     * or ACC_ALL.
     *
     * Return 0 if the specified access is allowed otherwise
     * return -1.
     */

    struct stat stbuf;
    int mode_r, mode_w, mode_x;
    char buf[MAXPATHNAME];
    uid_t uid, euid;
    gid_t gid, egid;
    

    if (stat(pathname, &stbuf) == 0) {
        if ((stbuf.st_mode & S_IFMT) == S_IFLNK) {
            /* Follow the link to the actual file */
            memset(buf, '\0', MAXPATHNAME);
            if (readlink(pathname, buf, MAXPATHNAME) > 0) {
                return accessx(buf, mode, who);
            } else {
                return -1;
            }
        }
    }

    /*
     * Break apart the bit-fields
     */
    
    mode_r = (((mode & R_OK) == R_OK) ? 1 : 0);
    mode_w = (((mode & W_OK) == W_OK) ? 1 : 0);
    mode_x = (((mode & X_OK) == X_OK) ? 1 : 0);

    /*
     * Find out who I am
     */

    uid = getuid();
    euid = geteuid();
    gid = getgid();
    egid = getegid();

    /*
     * Permission masks
     *
     * S_IRUSR, S_IWUSR, S_IXUSR
     * S_IRGRP, S_IWGRP, S_IXGRP
     * S_IROTH, S_IWOTH, S_IXOTH
     *
     */
     
    switch (who) {
        case ACC_SELF:
            /*
             * Determine if access is permitted for the current process.
             * The effective user and group IDs and the concurrent group
             * set are used in the calculation.
             */
            if (stbuf.st_uid == euid || euid == 0) {
                if (mode_r) {
                    if ((stbuf.st_mode & S_IRUSR) == 0) {
                        return -1;
                    }
                }
                if (mode_w) {
                    if ((stbuf.st_mode & S_IWUSR) == 0) {
                        return -1;
                    }
                }
                if (mode_x) {
                    if ((stbuf.st_mode & S_IXUSR) == 0) {
                        return -1;
                    }
                }
                return 0;       /* Access is permitted */
            } else if (stbuf.st_gid == egid) {
            }
            break;
            
        case ACC_INVOKER:
            /*
             * The real user and group IDs, the concurrent group set
             * and the process privilege are used to determine if
             * access is permitted.
             */

            if (stbuf.st_uid == uid || euid == 0) {
                if (mode_r) {
                    if ((stbuf.st_mode & S_IRUSR) == 0) {
                        return -1;
                    }
                }
                if (mode_w) {
                    if ((stbuf.st_mode & S_IWUSR) == 0) {
                        return -1;
                    }
                }
                if (mode_x) {
                    if ((stbuf.st_mode & S_IXUSR) == 0) {
                        return -1;
                    }
                }
                return 0;       /* Access is permitted */
            } else if (stbuf.st_gid == egid) {
                if (mode_r) {
                    if ((stbuf.st_mode & S_IRGRP) == 0) {
                        return -1;
                    }
                }
                if (mode_w) {
                    if ((stbuf.st_mode & S_IWGRP) == 0) {
                        return -1;
                    }
                }
                if (mode_x) {
                    if ((stbuf.st_mode & S_IXGRP) == 0) {
                        return -1;
                    }
                }                
            }
            break;
            
        case ACC_OTHERS:
            /*
             * Determines if access is allowed for any user other than the
             * object owner.
             */

            if (mode_r) {
                if (((stbuf.st_mode & S_IRGRP) == 0) &&
                    ((stbuf.st_mode & S_IROTH) == 0)) {
                    return -1;
                }
            }
            if (mode_w) {
                if (((stbuf.st_mode & S_IWGRP) == 0) &&
                    ((stbuf.st_mode & S_IWOTH) == 0)) {
                    return -1;
                }
            }
            if (mode_x) {
                if (((stbuf.st_mode & S_IXGRP) == 0) &&
                    ((stbuf.st_mode & S_IXOTH) == 0)) {
                    return -1;
                }
            }
            return 0;       /* access is permitted */

                
        case ACC_ALL:
            /*
             * Determine if access is allowed for all users
             */
                
            if (mode_r) {
                if (((stbuf.st_mode & S_IRUSR) == 0) ||
                    ((stbuf.st_mode & S_IRGRP) == 0) ||
                    (stbuf.st_mode & S_IROTH) == 0) {
                    return -1;
                }
            }
            if (mode_w) {
                if (((stbuf.st_mode & S_IWUSR) == 0) ||
                    ((stbuf.st_mode & S_IWGRP) == 0) ||
                    (stbuf.st_mode & S_IWOTH) == 0) {
                    return -1;
                }
            }
            if (mode_x) {
                if (((stbuf.st_mode & S_IXUSR) == 0) ||
                    ((stbuf.st_mode & S_IXGRP) == 0) ||
                    (stbuf.st_mode & S_IXOTH) == 0) {
                    return -1;
                }
            }
            return 0;       /* access is permitted */
    }

    /* Not reached */
}

#endif /* _HAVE_ACCESSX */
